"""
Main entry point for Superset commands.
"""
from typing import Any

import click
from yarl import URL

from preset_cli.auth.main import UsernamePasswordAuth
from preset_cli.cli.superset.export import (
    export_assets,
    export_ownership,
    export_rls,
    export_users,
)
from preset_cli.cli.superset.import_ import import_ownership, import_rls
from preset_cli.cli.superset.sql import sql
from preset_cli.cli.superset.sync.main import sync
from preset_cli.cli.superset.sync.native.command import native


@click.group()
@click.argument("instance")
@click.option("-u", "--username", default="admin", help="Username")
@click.option(
    "-p",
    "--password",
    prompt=True,
    prompt_required=False,
    default="admin",
    hide_input=True,
    help="Password (leave empty for prompt)",
)
@click.pass_context
def superset_cli(
    ctx: click.core.Context,
    instance: str,
    username: str = "admin",
    password: str = "admin",
):
    """
    An Apache Superset CLI.
    """
    ctx.ensure_object(dict)

    ctx.obj["INSTANCE"] = instance

    # allow a custom authenticator to be passed via the context
    if "AUTH" not in ctx.obj:
        ctx.obj["AUTH"] = UsernamePasswordAuth(URL(instance), username, password)


superset_cli.add_command(sql)
superset_cli.add_command(sync)
superset_cli.add_command(export_assets)
superset_cli.add_command(export_assets, name="export")  # for backwards compatibility
superset_cli.add_command(export_users)
superset_cli.add_command(export_rls)
superset_cli.add_command(export_ownership)
superset_cli.add_command(import_rls)
superset_cli.add_command(import_ownership)
superset_cli.add_command(native, name="import-assets")


@click.group()
@click.pass_context
def superset(ctx: click.core.Context) -> None:
    """
    Send commands to one or more Superset instances.
    """
    ctx.ensure_object(dict)


def mutate_commands(source: click.core.Group, target: click.core.Group) -> None:
    """
    Programmatically modify commands so they work with workspaces.
    """
    for name, command in source.commands.items():

        if isinstance(command, click.core.Group):

            @click.group()
            @click.pass_context
            def new_group(
                ctx: click.core.Context, *args: Any, command=command, **kwargs: Any
            ) -> None:
                ctx.invoke(command, *args, **kwargs)

            mutate_commands(command, new_group)
            new_group.params = command.params[:]
            target.add_command(new_group, name)

        else:

            @click.command()
            @click.pass_context
            def new_command(
                ctx: click.core.Context, *args: Any, command=command, **kwargs: Any
            ) -> None:
                for instance in ctx.obj["WORKSPACES"]:
                    click.echo(f"\n{instance}")
                    ctx.obj["INSTANCE"] = instance
                    ctx.invoke(command, *args, **kwargs)

            new_command.params = command.params[:]
            target.add_command(new_command, name)


mutate_commands(superset_cli, superset)



"""
A simple client for interacting with the dbt API.

References:

    - https://docs.getdbt.com/dbt-cloud/api-v2
    - https://docs.getdbt.com/docs/dbt-cloud/dbt-cloud-api/metadata/schema/metadata-schema-seeds
    - https://github.com/dbt-labs/dbt-cloud-openapi-spec/blob/master/openapi-v3.yaml
"""

# pylint: disable=invalid-name, too-few-public-methods

from enum import Enum
from typing import Any, Dict, List, Type, TypedDict

from marshmallow import INCLUDE, Schema, fields
from python_graphql_client import GraphqlClient
from yarl import URL

from preset_cli import __version__
from preset_cli.auth.main import Auth

REST_ENDPOINT = URL("https://cloud.getdbt.com/")
GRAPHQL_ENDPOINT = URL("https://metadata.cloud.getdbt.com/graphql")


class PostelSchema(Schema):
    """
    Be liberal in what you accept, and conservative in what you send.

    A schema that allows unknown fields. This way if they API returns new fields that
    the client is not expecting no errors will be thrown when validating the payload.
    """

    class Meta:
        """
        Allow unknown fields.
        """

        unknown = INCLUDE


def PostelEnumField(enum: Type[Enum], *args: Any, **kwargs: Any) -> fields.Field:
    """
    Lenient replacement for ``EnumField``.

    This allows us to keep track of the enums expected in a field, while still
    accepting any unexpected new values that are introduced.
    """
    if issubclass(enum, str):
        return fields.String(*args, **kwargs)

    if issubclass(enum, int):
        return fields.Integer(*args, **kwargs)

    return fields.Raw(*args, **kwargs)


class DBTAccountPlan(str, Enum):
    """
    Plan for the account.
    """

    developer = "developer"
    team = "team"
    enterprise = "enterprise"


class DBTUserPlan(str, Enum):
    """
    Plan for the user.
    """

    free = "free"
    trial = "trial"
    enterprise = "enterprise"
    developer = "developer"
    team = "team"
    cancelled = "cancelled"


class EnterpriseAuthentication(str, Enum):
    """
    Authentication methods for enterprise.
    """

    none = "none"
    okta = "okta"
    azure_ad = "azure_ad"
    gsuite = "gsuite"


class GitAuthLevel(str, Enum):
    """
    Indicates the git provider authentication level for this user
    """

    personal = "personal"
    team = "team"


class AccountSchema(PostelSchema):
    """
    Schema for a dbt account.
    """

    id = fields.Integer()
    name = fields.String()
    plan = PostelEnumField(DBTAccountPlan)
    pending_cancel = fields.Boolean()
    state = fields.Integer()
    developer_seats = fields.Integer()
    read_only_seats = fields.Integer()
    run_slots = fields.Integer()
    created_at = fields.DateTime()
    updated_at = fields.DateTime()


class LicenseType(str, Enum):
    """
    Types of licenses.
    """

    developer = "developer"
    read_only = "read_only"


class GroupPermission(str, Enum):
    """
    Group permissions.
    """

    owner = "owner"
    member = "member"
    account_admin = "account_admin"
    admin = "admin"
    database_admin = "database_admin"
    git_admin = "git_admin"
    team_admin = "team_admin"
    job_admin = "job_admin"
    job_viewer = "job_viewer"
    analyst = "analyst"
    developer = "developer"
    stakeholder = "stakeholder"
    readonly = "readonly"
    project_creator = "project_creator"
    account_viewer = "account_viewer"
    metadata_only = "metadata_only"
    webhooks_only = "webhooks_only"


class GroupPermissionSchema(PostelSchema):
    """
    Schema for group permissions.
    """

    id = fields.String()
    account_id = fields.String()
    project_id = fields.String(allow_none=True)
    all_projects = fields.Boolean()
    permission_set = PostelEnumField(GroupPermission)
    permission_level = fields.Integer(allow_none=True)
    state = fields.Integer()


class GroupSchema(PostelSchema):
    """
    A dbt group.
    """

    id = fields.Integer()
    account_id = fields.Integer()
    name = fields.String()
    state = fields.Integer()
    assign_by_default = fields.Boolean()
    group_permissions = fields.List(fields.Nested(GroupPermissionSchema))


class PermissionStatement(str, Enum):
    """
    Permission types.
    """

    billing_read = "billing_read"
    billing_write = "billing_write"
    invitations_send = "invitations_send"
    invitations_modify = "invitations_modify"
    invitations_read = "invitations_read"
    members_read = "members_read"
    members_write = "members_write"
    groups_read = "groups_read"
    groups_write = "groups_write"
    license_read = "license_read"
    license_allocate = "license_allocate"
    projects_read = "projects_read"
    projects_develop = "projects_develop"
    projects_write = "projects_write"
    projects_create = "projects_create"
    projects_delete = "projects_delete"
    environments_read = "environments_read"
    environments_write = "environments_write"
    develop_access = "develop_access"
    dbt_adapters_read = "dbt_adapters_read"
    dbt_adapters_write = "dbt_adapters_write"
    credentials_read = "credentials_read"
    credentials_write = "credentials_write"
    connections_read = "connections_read"
    connections_write = "connections_write"
    jobs_read = "jobs_read"
    jobs_write = "jobs_write"
    repositories_read = "repositories_read"
    repositories_write = "repositories_write"
    runs_trigger = "runs_trigger"
    runs_write = "runs_write"
    runs_read = "runs_read"
    permissions_write = "permissions_write"
    permissions_read = "permissions_read"
    account_settings_write = "account_settings_write"
    account_settings_read = "account_settings_read"
    auth_provider_write = "auth_provider_write"
    auth_provider_read = "auth_provider_read"
    service_tokens_write = "service_tokens_write"
    service_tokens_read = "service_tokens_read"
    metadata_read = "metadata_read"
    webhooks_write = "webhooks_write"
    custom_environment_variables_read = "custom_environment_variables_read"
    custom_environment_variables_write = "custom_environment_variables_write"
    audit_log_read = "audit_log_read"


class PermissionStatementSchema(PostelSchema):
    """
    Permission statements.
    """

    permission = PostelEnumField(PermissionStatement)
    target_resource = fields.Integer()
    all_resources = fields.Boolean()


class UserLicenseSchema(PostelSchema):
    """
    Schema for user permissions.
    """

    id = fields.Integer()
    license_type = PostelEnumField(LicenseType)
    user_id = fields.Integer()
    account_id = fields.Integer()
    state = fields.Integer()
    groups = fields.List(fields.Nested(GroupSchema))
    permission_statements = fields.List(fields.Nested(PermissionStatementSchema))


class UserSchema(PostelSchema):
    """
    Schema for a dbt user.
    """

    id = fields.Integer()
    state = fields.Integer()
    name = fields.String()
    lock_reason = fields.String(allow_none=True)
    unlock_if_subscription_renewed = fields.Boolean()
    plan = PostelEnumField(DBTUserPlan)
    pending_cancel = fields.Boolean()
    run_slots = fields.Integer()
    developer_seats = fields.Integer()
    read_only_seats = fields.Integer()
    queue_limit = fields.Integer()
    pod_memory_request_mebibytes = fields.Integer()
    run_duration_limit_seconds = fields.Integer()
    enterprise_authentication_method = PostelEnumField(
        EnterpriseAuthentication,
        allow_none=True,
    )
    enterprise_login_slug = fields.String(allow_none=True)
    enterprise_unique_identifier = fields.String(allow_none=True)
    billing_email_address = fields.String(allow_none=True)
    locked = fields.Boolean()
    unlocked_at = fields.DateTime()
    created_at = fields.DateTime()
    updated_at = fields.DateTime()
    starter_repo_url = fields.String(allow_none=True)
    sso_reauth = fields.Boolean()
    git_auth_level = PostelEnumField(GitAuthLevel, allow_none=True)
    identifier = fields.String()
    docs_job_id = fields.Integer(allow_none=True)
    freshness_job_id = fields.Integer(allow_none=True)
    docs_job = fields.String(allow_none=True)
    freshness_job = fields.String(allow_none=True)
    enterprise_login_url = fields.String(allow_none=True)
    permissions = fields.Nested(UserLicenseSchema)

    # not present in the spec
    develop_file_system = fields.Boolean()
    force_sso = fields.Boolean()


class ConnectionType(str, Enum):
    """
    Connection types.
    """

    postgres = "postgres"
    redshift = "redshift"
    snowflake = "snowflake"
    bigquery = "bigquery"
    adapter = "adapter"


class ConnectionSchema(PostelSchema):
    """
    Schema for connection information.
    """

    id = fields.Integer()
    account_id = fields.Integer()
    project_id = fields.Integer()
    name = fields.String()
    type = PostelEnumField(ConnectionType)
    state = fields.Integer()
    created_by_id = fields.Integer(allow_none=True)
    created_by_service_token_id = fields.Integer(allow_none=True)
    created_at = fields.DateTime()
    updated_at = fields.DateTime()
    details = fields.Raw()


class GitCloneStrategy(str, Enum):
    """
    Types of strategies for cloning git.
    """

    azure_active_directory_app = "azure_active_directory_app"
    deploy_key = "deploy_key"
    deploy_token = "deploy_token"
    github_app = "github_app"
    git_token = "git_token"


class DeployKeySchema(PostelSchema):
    """
    Schema for a deployment key.
    """

    id = fields.Integer()
    account_id = fields.Integer()
    state = fields.Integer()
    public_key = fields.String()


class RepositorySchema(PostelSchema):
    """
    Schema for a repository.
    """

    id = fields.Integer()
    account_id = fields.Integer()
    remote_url = fields.String()
    remote_backend = fields.String()
    git_clone_strategy = PostelEnumField(GitCloneStrategy)
    deploy_key_id = fields.Integer()
    github_installation_id = fields.Integer()
    state = fields.Integer()
    created_at = fields.DateTime()
    updated_at = fields.DateTime()

    # not present in the spec
    full_name = fields.String()
    repository_credentials_id = fields.Integer(allow_none=True)
    gitlab = fields.String(allow_none=True)
    name = fields.String()
    pull_request_url_template = fields.String()
    git_provider_id = fields.Integer()
    git_provider = fields.String(allow_none=True)
    project_id = fields.Integer()
    deploy_key = fields.Nested(DeployKeySchema)
    github_repo = fields.String()


class ProjectSchema(PostelSchema):
    """
    Schema for a dbt project.
    """

    id = fields.Integer()
    account_id = fields.Integer()
    connection = fields.Nested(ConnectionSchema)
    connection_id = fields.Integer()
    dbt_project_subdirectory = fields.String(allow_none=True)
    name = fields.String()
    repository = fields.Nested(RepositorySchema)
    repository_id = fields.Integer()
    state = fields.Integer()
    created_at = fields.DateTime()
    updated_at = fields.DateTime()

    # not present in the spec
    group_permissions = fields.List(fields.Nested(GroupPermissionSchema))
    docs_job = fields.String(allow_none=True)
    docs_job_id = fields.Integer(allow_none=True)
    freshness_job_id = fields.Integer(allow_none=True)
    freshness_job = fields.String(allow_none=True)
    skipped_setup = fields.Boolean()


class TriggerSchema(PostelSchema):
    """
    Schema for a job trigger.
    """

    github_webhook = fields.Boolean(required=True)
    git_provider_webhook = fields.Boolean()
    schedule = fields.Boolean(required=True)
    custom_branch_only = fields.Boolean()


class SettingsSchema(PostelSchema):
    """
    Schema for job settings.
    """

    threads = fields.Integer(required=True)
    target_name = fields.String(required=True)


class DateType(str, Enum):
    """
    Types of date.
    """

    every_day = "every_day"
    days_of_week = "days_of_week"
    custom_cron = "custom_cron"


class DateSchema(PostelSchema):
    """
    Schema for a date.
    """

    type = PostelEnumField(DateType, required=True)
    days = fields.List(fields.Integer(), allow_none=True)
    cron = fields.String(allow_none=True)


class TimeType(str, Enum):
    """
    Types of time.
    """

    every_hour = "every_hour"
    at_exact_hours = "at_exact_hours"


class TimeSchema(PostelSchema):
    """
    Schema for a time.
    """

    type = PostelEnumField(TimeType, required=True)
    interval = fields.Integer(allow_none=True)
    hours = fields.List(fields.Integer(), allow_none=True)


class ScheduleSchema(PostelSchema):
    """
    Schema for a job schedule.
    """

    cron = fields.String()
    date = fields.Nested(DateSchema, required=True)
    time = fields.Nested(TimeSchema, required=True)


class ExecutionSchema(PostelSchema):
    """
    Schema for a job execution.
    """

    timeout_seconds = fields.Integer()


class JobSchema(PostelSchema):
    """
    Schema for a dbt job.
    """

    id = fields.Integer(allow_none=True)
    account_id = fields.Integer()
    project_id = fields.Integer()
    environment_id = fields.Integer()
    name = fields.String()
    dbt_version = fields.String(allow_none=True)
    triggers = fields.Nested(TriggerSchema)
    execute_steps = fields.List(fields.String())
    settings = fields.Nested(SettingsSchema)
    state = fields.Integer()
    generate_docs = fields.Boolean()
    schedule = fields.Nested(ScheduleSchema)

    # not present in the spec
    lifecycle_webhooks_url = fields.String(allow_none=True)
    cron_humanized = fields.String()
    created_at = fields.DateTime()
    next_run = fields.DateTime()
    lifecycle_webhooks = fields.Boolean()
    next_run_humanized = fields.String()
    deferring_job_definition_id = fields.Integer(allow_none=True)
    deactivated = fields.Boolean()
    is_deferrable = fields.Boolean()
    updated_at = fields.DateTime()
    execution = fields.Nested(ExecutionSchema)
    run_failure_count = fields.Integer()
    run_generate_sources = fields.Boolean()
    generate_sources = fields.Boolean()


class ModelSchema(PostelSchema):
    """
    Schema for a model.
    """

    depends_on = fields.List(fields.String(), data_key="dependsOn")
    children = fields.List(fields.String(), data_key="childrenL1")
    database = fields.String()
    schema = fields.String()
    description = fields.String()
    meta = fields.Raw()
    name = fields.String()
    unique_id = fields.String(data_key="uniqueId")
    tags = fields.List(fields.String())


class FilterSchema(PostelSchema):
    """
    Schema for a metric filter.
    """

    field = fields.String()
    operator = fields.String()
    value = fields.String()


class MetricSchema(PostelSchema):
    """
    Schema for a metric.
    """

    depends_on = fields.List(fields.String(), data_key="dependsOn")
    description = fields.String()
    filters = fields.List(fields.Nested(FilterSchema))
    meta = fields.Raw()
    name = fields.String()
    label = fields.String()
    sql = fields.String()
    type = fields.String()
    unique_id = fields.String(data_key="uniqueId")


class DataResponse(TypedDict):
    """
    Type for the GraphQL response.
    """

    data: Dict[str, Any]


class DBTClient:  # pylint: disable=too-few-public-methods

    """
    A client for the dbt API.
    """

    def __init__(self, auth: Auth):
        self.auth = auth
        self.auth.headers.update(
            {
                "User-Agent": "Preset CLI",
                "X-Client-Version": __version__,
            },
        )
        self.graphql_client = GraphqlClient(endpoint=GRAPHQL_ENDPOINT)
        self.baseurl = REST_ENDPOINT

    def execute(self, query: str, **variables: Any) -> DataResponse:
        """
        Run a GraphQL query.
        """
        return self.graphql_client.execute(
            query=query,
            variables=variables,
            headers=self.auth.get_headers(),
        )

    def get_accounts(self) -> List[AccountSchema]:
        """
        List all accounts.
        """
        session = self.auth.get_session()
        headers = self.auth.get_headers()
        response = session.get(self.baseurl / "api/v2/accounts/", headers=headers)

        payload = response.json()

        account_schema = AccountSchema()
        return [account_schema.load(row) for row in payload["data"]]

    def get_projects(self, account_id: int) -> List[ProjectSchema]:
        """
        List all projects.
        """
        session = self.auth.get_session()
        headers = self.auth.get_headers()
        response = session.get(
            self.baseurl / "api/v2/accounts" / str(account_id) / "projects/",
            headers=headers,
        )

        payload = response.json()

        project_schema = ProjectSchema()
        projects = [project_schema.load(project) for project in payload["data"]]

        return projects

    def get_jobs(self, account_id: int) -> List[JobSchema]:
        """
        List all jobs.
        """
        session = self.auth.get_session()
        headers = self.auth.get_headers()
        response = session.get(
            self.baseurl / "api/v2/accounts" / str(account_id) / "jobs/",
            headers=headers,
        )

        payload = response.json()

        job_schema = JobSchema()
        jobs = [job_schema.load(job) for job in payload["data"]]

        return jobs

    def get_models(self, job_id: int) -> List[ModelSchema]:
        """
        Fetch all available models.
        """
        query = """
            query ($jobId: Int!) {
                models(jobId: $jobId) {
                    uniqueId
                    dependsOn
                    childrenL1
                    name
                    database
                    schema
                    description
                    meta
                    tags
                }
            }
        """
        payload = self.execute(query, jobId=job_id)

        model_schema = ModelSchema()
        models = [model_schema.load(model) for model in payload["data"]["models"]]

        return models

    def get_metrics(self, job_id: int) -> List[Any]:
        """
        Fetch all available metrics.
        """
        query = """
            query ($jobId: Int!) {
                metrics(jobId: $jobId) {
                    uniqueId
                    name
                    label
                    type
                    sql
                    filters {
                        field
                        operator
                        value
                    }
                    dependsOn
                    description
                    meta
                }
            }
        """
        payload = self.execute(query, jobId=job_id)

        metric_schema = MetricSchema()
        metrics = [metric_schema.load(metric) for metric in payload["data"]["metrics"]]

        return metrics

    def get_database_name(self, job_id: int) -> str:
        """
        Return the database name.

        This is done by querying all models in a job. As far as I know there should be
        only one database associated with them.
        """
        models = self.get_models(job_id)
        if not models:
            raise Exception("No models found, can't determine database name")

        databases = {model["database"] for model in models}
        if len(databases) > 1:
            raise Exception("Multiple databases found")

        return databases.pop()



"""
Tests for the dbt import command.
"""
# pylint: disable=invalid-name

import os
from pathlib import Path

import pytest
from click.testing import CliRunner
from pyfakefs.fake_filesystem import FakeFilesystem
from pytest_mock import MockerFixture

from preset_cli.cli.superset.main import superset_cli
from preset_cli.exceptions import DatabaseNotFoundError

dirname, _ = os.path.split(os.path.abspath(__file__))
with open(os.path.join(dirname, "manifest.json"), encoding="utf-8") as fp:
    manifest_contents = fp.read()


def test_dbt_core(mocker: MockerFixture, fs: FakeFilesystem) -> None:
    """
    Test the ``dbt-core`` command.
    """
    root = Path("/path/to/root")
    fs.create_dir(root)
    manifest = root / "default/target/manifest.json"
    fs.create_file(manifest, contents=manifest_contents)
    profiles = root / ".dbt/profiles.yml"
    fs.create_file(profiles)
    exposures = root / "models/exposures.yml"
    fs.create_file(exposures)

    SupersetClient = mocker.patch(
        "preset_cli.cli.superset.sync.dbt.command.SupersetClient",
    )
    client = SupersetClient()
    mocker.patch("preset_cli.cli.superset.main.UsernamePasswordAuth")
    sync_database = mocker.patch(
        "preset_cli.cli.superset.sync.dbt.command.sync_database",
    )
    sync_datasets = mocker.patch(
        "preset_cli.cli.superset.sync.dbt.command.sync_datasets",
    )
    sync_exposures = mocker.patch(
        "preset_cli.cli.superset.sync.dbt.command.sync_exposures",
    )

    runner = CliRunner()
    result = runner.invoke(
        superset_cli,
        [
            "https://superset.example.org/",
            "sync",
            "dbt-core",
            str(manifest),
            "--profiles",
            str(profiles),
            "--exposures",
            str(exposures),
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 0
    sync_database.assert_called_with(
        client,
        profiles,
        "default",
        "dev",
        False,
        False,
        "",
    )
    models = [
        {
            "database": "examples_dev",
            "description": "",
            "meta": {},
            "name": "messages_channels",
            "schema": "public",
            "unique_id": "model.superset_examples.messages_channels",
            "tags": [],
        },
    ]
    metrics = [
        {
            "depends_on": ["model.superset_examples.messages_channels"],
            "description": "",
            "filters": [],
            "label": "",
            "meta": {},
            "name": "cnt",
            "sql": "*",
            "type": "count",
        },
    ]
    sync_datasets.assert_called_with(
        client,
        models,
        metrics,
        sync_database(),
        False,
        "",
    )
    sync_exposures.assert_called_with(client, exposures, sync_datasets())


def test_dbt(mocker: MockerFixture, fs: FakeFilesystem) -> None:
    """
    Test the ``dbt`` command.

    Initially ``dbt-core`` was just ``dbt``. This aliases was added for backwards
    compatibility.
    """
    root = Path("/path/to/root")
    fs.create_dir(root)
    manifest = root / "default/target/manifest.json"
    fs.create_file(manifest, contents=manifest_contents)
    profiles = root / ".dbt/profiles.yml"
    fs.create_file(profiles)
    exposures = root / "models/exposures.yml"
    fs.create_file(exposures)

    SupersetClient = mocker.patch(
        "preset_cli.cli.superset.sync.dbt.command.SupersetClient",
    )
    client = SupersetClient()
    mocker.patch("preset_cli.cli.superset.main.UsernamePasswordAuth")
    sync_database = mocker.patch(
        "preset_cli.cli.superset.sync.dbt.command.sync_database",
    )
    sync_datasets = mocker.patch(
        "preset_cli.cli.superset.sync.dbt.command.sync_datasets",
    )
    sync_exposures = mocker.patch(
        "preset_cli.cli.superset.sync.dbt.command.sync_exposures",
    )

    runner = CliRunner()
    result = runner.invoke(
        superset_cli,
        [
            "https://superset.example.org/",
            "sync",
            "dbt",
            str(manifest),
            "--profiles",
            str(profiles),
            "--exposures",
            str(exposures),
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 0
    sync_database.assert_called_with(
        client,
        profiles,
        "default",
        "dev",
        False,
        False,
        "",
    )
    models = [
        {
            "database": "examples_dev",
            "description": "",
            "meta": {},
            "name": "messages_channels",
            "schema": "public",
            "unique_id": "model.superset_examples.messages_channels",
            "tags": [],
        },
    ]
    metrics = [
        {
            "depends_on": ["model.superset_examples.messages_channels"],
            "description": "",
            "filters": [],
            "label": "",
            "meta": {},
            "name": "cnt",
            "sql": "*",
            "type": "count",
        },
    ]
    sync_datasets.assert_called_with(
        client,
        models,
        metrics,
        sync_database(),
        False,
        "",
    )
    sync_exposures.assert_called_with(client, exposures, sync_datasets())


def test_dbt_core_no_exposures(mocker: MockerFixture, fs: FakeFilesystem) -> None:
    """
    Test the ``dbt-core`` command when no exposures file is passed.
    """
    root = Path("/path/to/root")
    fs.create_dir(root)
    manifest = root / "default/target/manifest.json"
    fs.create_file(manifest, contents=manifest_contents)
    profiles = root / ".dbt/profiles.yml"
    fs.create_file(profiles)

    mocker.patch("preset_cli.cli.superset.sync.dbt.command.SupersetClient")
    mocker.patch("preset_cli.cli.superset.main.UsernamePasswordAuth")
    mocker.patch("preset_cli.cli.superset.sync.dbt.command.sync_database")
    mocker.patch("preset_cli.cli.superset.sync.dbt.command.sync_datasets")
    sync_exposures = mocker.patch(
        "preset_cli.cli.superset.sync.dbt.command.sync_exposures",
    )

    runner = CliRunner()
    result = runner.invoke(
        superset_cli,
        [
            "https://superset.example.org/",
            "sync",
            "dbt-core",
            str(manifest),
            "--profiles",
            str(profiles),
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 0
    sync_exposures.assert_not_called()


def test_dbt_core_default_profile(mocker: MockerFixture, fs: FakeFilesystem) -> None:
    """
    Test the ``dbt-core`` command when the profile is not passed
    """
    root = Path("/path/to/root")
    fs.create_dir(root)
    manifest = root / "default/target/manifest.json"
    fs.create_file(manifest, contents=manifest_contents)
    profiles = root / ".dbt/profiles.yml"
    fs.create_file(profiles)
    exposures = root / "models/exposures.yml"
    fs.create_file(exposures)

    SupersetClient = mocker.patch(
        "preset_cli.cli.superset.sync.dbt.command.SupersetClient",
    )
    client = SupersetClient()
    mocker.patch("preset_cli.cli.superset.main.UsernamePasswordAuth")
    sync_database = mocker.patch(
        "preset_cli.cli.superset.sync.dbt.command.sync_database",
    )
    mocker.patch("preset_cli.cli.superset.sync.dbt.command.sync_datasets")
    mocker.patch("preset_cli.cli.superset.sync.dbt.command.sync_exposures")
    # pylint: disable=redefined-outer-name
    os = mocker.patch("preset_cli.cli.superset.sync.dbt.command.os")
    os.path.expanduser.return_value = str(profiles)

    runner = CliRunner()
    result = runner.invoke(
        superset_cli,
        [
            "https://superset.example.org/",
            "sync",
            "dbt-core",
            str(manifest),
            "--exposures",
            str(exposures),
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 0
    sync_database.assert_called_with(
        client,
        profiles,
        "default",
        "dev",
        False,
        False,
        "",
    )


def test_dbt_core_no_database(mocker: MockerFixture, fs: FakeFilesystem) -> None:
    """
    Test the ``dbt-core`` command when no database is found and ``--import-db`` not passed.
    """
    root = Path("/path/to/root")
    fs.create_dir(root)
    manifest = root / "default/target/manifest.json"
    fs.create_file(manifest, contents=manifest_contents)
    profiles = root / ".dbt/profiles.yml"
    fs.create_file(profiles)
    exposures = root / "models/exposures.yml"
    fs.create_file(exposures)

    mocker.patch("preset_cli.cli.superset.sync.dbt.command.SupersetClient")
    mocker.patch("preset_cli.cli.superset.main.UsernamePasswordAuth")
    mocker.patch(
        "preset_cli.cli.superset.sync.dbt.command.sync_database",
        side_effect=DatabaseNotFoundError(),
    )

    runner = CliRunner()
    result = runner.invoke(
        superset_cli,
        [
            "https://superset.example.org/",
            "sync",
            "dbt-core",
            str(manifest),
            "--profiles",
            str(profiles),
            "--exposures",
            str(exposures),
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 0
    assert result.output == "No database was found, pass --import-db to create\n"


def test_dbt_cloud(mocker: MockerFixture) -> None:
    """
    Test the ``dbt-cloud`` command.
    """
    SupersetClient = mocker.patch(
        "preset_cli.cli.superset.sync.dbt.command.SupersetClient",
    )
    superset_client = SupersetClient()
    mocker.patch("preset_cli.cli.superset.main.UsernamePasswordAuth")
    DBTClient = mocker.patch(
        "preset_cli.cli.superset.sync.dbt.command.DBTClient",
    )
    dbt_client = DBTClient()
    sync_datasets = mocker.patch(
        "preset_cli.cli.superset.sync.dbt.command.sync_datasets",
    )
    models = [
        {
            "database": "examples_dev",
            "description": "",
            "meta": {},
            "name": "messages_channels",
            "schema": "public",
            "unique_id": "model.superset_examples.messages_channels",
        },
    ]
    dbt_client.get_models.return_value = models
    metrics = [
        {
            "depends_on": ["model.superset_examples.messages_channels"],
            "description": "",
            "filters": [],
            "label": "",
            "meta": {},
            "name": "cnt",
            "sql": "*",
            "type": "count",
        },
    ]
    dbt_client.get_metrics.return_value = metrics
    database = mocker.MagicMock()
    superset_client.get_databases.return_value = [database]

    runner = CliRunner()
    result = runner.invoke(
        superset_cli,
        [
            "https://superset.example.org/",
            "sync",
            "dbt-cloud",
            "XXX",
            "123",
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 0
    sync_datasets.assert_called_with(
        superset_client,
        models,
        metrics,
        database,
        False,
        "",
    )


def test_dbt_cloud_no_database(mocker: MockerFixture) -> None:
    """
    Test the ``dbt-cloud`` command when no database is found.
    """
    SupersetClient = mocker.patch(
        "preset_cli.cli.superset.sync.dbt.command.SupersetClient",
    )
    superset_client = SupersetClient()
    mocker.patch("preset_cli.cli.superset.main.UsernamePasswordAuth")
    DBTClient = mocker.patch(
        "preset_cli.cli.superset.sync.dbt.command.DBTClient",
    )
    dbt_client = DBTClient()
    dbt_client.get_database_name.return_value = "my_db"
    superset_client.get_databases.return_value = []

    runner = CliRunner()
    result = runner.invoke(
        superset_cli,
        [
            "https://superset.example.org/",
            "sync",
            "dbt-cloud",
            "XXX",
            "123",
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 0
    assert result.output == 'No database named "my_db" was found\n'


def test_dbt_cloud_multiple_databases(mocker: MockerFixture) -> None:
    """
    Test the ``dbt-cloud`` command when multiple databases are found.

    This should never happen, since Supersret has a uniqueness contraint on the table
    name. Nevertheless, test this for completeness.
    """
    SupersetClient = mocker.patch(
        "preset_cli.cli.superset.sync.dbt.command.SupersetClient",
    )
    superset_client = SupersetClient()
    mocker.patch("preset_cli.cli.superset.main.UsernamePasswordAuth")
    DBTClient = mocker.patch(
        "preset_cli.cli.superset.sync.dbt.command.DBTClient",
    )
    dbt_client = DBTClient()
    dbt_client.get_database_name.return_value = "my_db"
    superset_client.get_databases.return_value = [
        mocker.MagicMock(),
        mocker.MagicMock(),
    ]

    runner = CliRunner()
    with pytest.raises(Exception) as excinfo:
        runner.invoke(
            superset_cli,
            [
                "https://superset.example.org/",
                "sync",
                "dbt-cloud",
                "XXX",
                "123",
            ],
            catch_exceptions=False,
        )
    assert str(excinfo.value) == "More than one database with the same name found"



"""
Main entry point for the CLI.
"""

import getpass
import logging
import sys
import webbrowser
from typing import List, Optional, cast

import click
import yaml
from yarl import URL

from preset_cli.api.clients.preset import PresetClient
from preset_cli.auth.jwt import JWTAuth
from preset_cli.auth.lib import (
    get_access_token,
    get_credentials_path,
    store_credentials,
)
from preset_cli.auth.main import Auth
from preset_cli.cli.superset.main import superset
from preset_cli.lib import setup_logging

_logger = logging.getLogger(__name__)


def split_comma(  # pylint: disable=unused-argument
    ctx: click.core.Context,
    param: str,
    value: Optional[str],
) -> List[str]:
    """
    Split CLI option into multiple values.
    """
    if value is None:
        return []

    return [option.strip() for option in value.split(",")]


def get_status_icon(status: str) -> str:
    """
    Return an icon (emoji) for a given status.
    """
    icons = {
        "READY": "✅",
        "LOADING_EXAMPLES": "📊",
        "CREATING_DB": "💾",
        "INITIALIZING_DB": "💾",
        "MIGRATING_DB": "🚧",
        "ROTATING_SECRETS": "🕵️",
        "UNKNOWN": "❓",
        "ERROR": "❗️",
        "UPGRADING": "⤴️",
    }
    return icons.get(status, "❓")


def parse_workspace_selection(selection: str, count: int) -> List[int]:
    """
    Parse a range of numbers.

        >>> parse_workspace_selection("1-4,7", 10)
        [1, 2, 3, 4, 7]

    """
    numbers: List[int] = []
    for part in selection.split(","):
        if "-" in part:
            if part[0] == "-":
                part = "1" + part
            if part[-1] == "-":
                part = part + str(count)
            start, end = [int(number) for number in part.split("-", 1)]
            if end > count:
                raise Exception(f"End {end} is greater than {count}")
            numbers.extend(range(start, end + 1))
        else:
            number = int(part)
            if number > count:
                raise Exception(f"Number {number} is greater than {count}")
            numbers.append(int(part))

    return numbers


@click.group()
@click.option("--baseurl", default="https://manage.app.preset.io/")
@click.option("--api-token", envvar="PRESET_API_TOKEN")
@click.option("--api-secret", envvar="PRESET_API_SECRET")
@click.option("--jwt-token", envvar="PRESET_JWT_TOKEN")
@click.option("--workspaces", callback=split_comma)
@click.option("--loglevel", default="INFO")
@click.pass_context
def preset_cli(  # pylint: disable=too-many-branches, too-many-locals, too-many-arguments
    ctx: click.core.Context,
    baseurl: str,
    api_token: Optional[str],
    api_secret: Optional[str],
    jwt_token: Optional[str],
    workspaces: List[str],
    loglevel: str,
) -> None:
    """
    A CLI for Preset.
    """
    setup_logging(loglevel)

    ctx.ensure_object(dict)

    # store manager URL for other commands
    ctx.obj["MANAGER_URL"] = manager_url = URL(baseurl)

    if jwt_token is None:
        if api_token is None or api_secret is None:
            # check for stored credentials
            credentials_path = get_credentials_path()
            if credentials_path.exists():
                try:
                    with open(credentials_path, encoding="utf-8") as input_:
                        credentials = yaml.load(input_, Loader=yaml.SafeLoader)
                    api_token = credentials["api_token"]
                    api_secret = credentials["api_secret"]
                except Exception:  # pylint: disable=broad-except
                    click.echo(
                        click.style(
                            "Couldn't read credentials",
                            fg="bright_red",
                        ),
                    )
                    sys.exit(1)
            else:
                click.echo(
                    "You need to specify a JWT token or an API key (name and secret)",
                )
                webbrowser.open(str(manager_url / "app/user"))
                api_token = input("API token: ")
                api_secret = getpass.getpass("API secret: ")
                store_credentials(api_token, api_secret, manager_url, credentials_path)

        api_token = cast(str, api_token)
        api_secret = cast(str, api_secret)
        try:
            jwt_token = get_access_token(manager_url, api_token, api_secret)
        except Exception:  # pylint: disable=broad-except
            jwt_token = None

    # store auth in context so it's used by the Superset SDK
    ctx.obj["AUTH"] = JWTAuth(jwt_token) if jwt_token else Auth()

    if not workspaces and ctx.invoked_subcommand == "superset":
        client = PresetClient(ctx.obj["MANAGER_URL"], ctx.obj["AUTH"])
        click.echo("Choose one or more workspaces (eg: 1-3,5,8-):")
        i = 0
        hostnames = []
        for team in client.get_teams():
            click.echo(f'\n# {team["title"]} #')
            for workspace in client.get_workspaces(team_name=team["name"]):
                status = get_status_icon(workspace["workspace_status"])
                click.echo(f'{status} ({i+1}) {workspace["title"]}')
                hostnames.append("https://" + workspace["hostname"])
                i += 1

        if i == 0:
            click.echo(
                click.style(
                    "No workspaces available",
                    fg="bright_red",
                ),
            )
            sys.exit(1)
        if i == 1:
            workspaces = hostnames

        while not workspaces:
            try:
                choices = parse_workspace_selection(input("> "), i)
                workspaces = [hostnames[choice - 1] for choice in choices]
            except Exception:  # pylint: disable=broad-except
                click.echo("Invalid choice")

    # store workspaces in order to invoke the command for each one
    ctx.obj["WORKSPACES"] = workspaces


@click.command()
@click.option("--baseurl", default="https://manage.app.preset.io/")
@click.option(
    "--overwrite",
    is_flag=True,
    default=False,
    help="Overwrite existing credentials",
)
@click.option(
    "--show",
    is_flag=True,
    default=False,
    help="Show existing credentials",
)
def auth(baseurl: str, overwrite: bool = False, show: bool = False) -> None:
    """
    Store credentials for auth.
    """
    credentials_path = get_credentials_path()

    if show:
        if not credentials_path.exists():
            click.echo(
                click.style(
                    (
                        f"The file {credentials_path} doesn't exist. "
                        "Run ``preset-cli auth`` to create it."
                    ),
                    fg="bright_red",
                ),
            )
            sys.exit(1)

        ruler = "=" * len(str(credentials_path))
        with open(credentials_path, encoding="utf-8") as input_:
            credentials = yaml.load(input_, Loader=yaml.SafeLoader)
            contents = yaml.dump(credentials)
        click.echo(f"{credentials_path}\n{ruler}\n{contents}")
        sys.exit(0)

    if credentials_path.exists() and not overwrite:
        click.echo(
            click.style(
                (
                    f"The file {credentials_path} already exists. "
                    "Pass --overwrite to replace it."
                ),
                fg="bright_red",
            ),
        )
        sys.exit(1)

    manager_url = URL(baseurl)
    click.echo(
        f"Please generate a new token at {manager_url} if you don't have one already",
    )
    webbrowser.open(str(manager_url / "app/user"))
    api_token = input("API token: ")
    api_secret = getpass.getpass("API secret: ")

    store_credentials(api_token, api_secret, manager_url, credentials_path)
    click.echo(f"Credentials stored in {credentials_path}")


preset_cli.add_command(auth)
preset_cli.add_command(superset)



"""
Tests for ``preset_cli.cli.superset.sync.dbt.exposures``.
"""
# pylint: disable=invalid-name

import copy
import json
from pathlib import Path
from typing import Any, Dict

import yaml
from pyfakefs.fake_filesystem import FakeFilesystem
from pytest_mock import MockerFixture
from yarl import URL

from preset_cli.cli.superset.sync.dbt.exposures import (
    get_chart_depends_on,
    get_dashboard_depends_on,
    sync_exposures,
)

dashboard_response: Dict[str, Any] = {
    "result": {
        "certification_details": None,
        "certified_by": None,
        "changed_by": {
            "first_name": "admin",
            "id": 1,
            "last_name": "admin",
            "username": "admin",
        },
        "changed_by_name": "admin admin",
        "changed_by_url": "/superset/profile/admin",
        "changed_on": "2022-03-27T13:23:25.741970",
        "changed_on_delta_humanized": "29 seconds ago",
        "charts": ["Example chart"],
        "css": None,
        "dashboard_title": "Example dashboard",
        "id": 12,
        "json_metadata": None,
        "owners": [
            {"first_name": "admin", "id": 1, "last_name": "admin", "username": "admin"},
        ],
        "position_json": None,
        "published": False,
        "roles": [],
        "slug": None,
        "thumbnail_url": "/api/v1/dashboard/12/thumbnail/1f7a46435e3ff1fefc4adc9e73aa0ae7/",
        "url": "/superset/dashboard/12/",
    },
}


datasets_response: Dict[str, Any] = {
    "result": [
        {
            "cache_timeout": None,
            "column_formats": {},
            "column_types": [1, 2],
            "columns": [
                {
                    "certification_details": None,
                    "certified_by": None,
                    "column_name": "name",
                    "description": None,
                    "expression": None,
                    "filterable": True,
                    "groupby": True,
                    "id": 842,
                    "is_certified": False,
                    "is_dttm": False,
                    "python_date_format": None,
                    "type": "VARCHAR(255)",
                    "type_generic": 1,
                    "verbose_name": None,
                    "warning_markdown": None,
                },
            ],
            "database": {
                "allow_multi_schema_metadata_fetch": False,
                "allows_cost_estimate": None,
                "allows_subquery": True,
                "allows_virtual_table_explore": True,
                "backend": "postgresql",
                "disable_data_preview": False,
                "explore_database_id": 3,
                "id": 3,
                "name": "superset_examples_dev",
            },
            "datasource_name": "messages_channels",
            "default_endpoint": None,
            "edit_url": "/tablemodelview/edit/27",
            "fetch_values_predicate": None,
            "filter_select": False,
            "filter_select_enabled": False,
            "granularity_sqla": [["ts", "ts"]],
            "health_check_message": None,
            "id": 27,
            "is_sqllab_view": False,
            "main_dttm_col": "ts",
            "metrics": [
                {
                    "certification_details": None,
                    "certified_by": None,
                    "d3format": None,
                    "description": "",
                    "expression": "count(*)",
                    "id": 35,
                    "is_certified": False,
                    "metric_name": "cnt",
                    "verbose_name": "count(*)",
                    "warning_markdown": None,
                    "warning_text": None,
                },
            ],
            "name": "public.messages_channels",
            "offset": 0,
            "order_by_choices": [
                ['["name", True]', "name [asc]"],
                ['["name", False]', "name [desc]"],
                ['["text", True]', "text [asc]"],
                ['["text", False]', "text [desc]"],
                ['["ts", True]', "ts [asc]"],
                ['["ts", False]', "ts [desc]"],
            ],
            "owners": [],
            "params": None,
            "perm": "[superset_examples_dev].[messages_channels](id:27)",
            "schema": "public",
            "select_star": "SELECT *\nFROM public.messages_channels\nLIMIT 100",
            "sql": None,
            "table_name": "messages_channels",
            "template_params": None,
            "time_grain_sqla": [
                [None, "Original value"],
                ["PT1S", "Second"],
                ["PT1M", "Minute"],
                ["PT1H", "Hour"],
                ["P1D", "Day"],
                ["P1W", "Week"],
                ["P1M", "Month"],
                ["P3M", "Quarter"],
                ["P1Y", "Year"],
            ],
            "type": "table",
            "uid": "27__table",
            "verbose_map": {"__timestamp": "Time", "cnt": "count(*)", "name": "name"},
        },
    ],
}

chart_response = {
    "description_columns": {},
    "id": 1,
    "label_columns": {
        "cache_timeout": "Cache Timeout",
        "certification_details": "Certification Details",
        "certified_by": "Certified By",
        "dashboards.dashboard_title": "Dashboards Dashboard Title",
        "dashboards.id": "Dashboards Id",
        "dashboards.json_metadata": "Dashboards Json Metadata",
        "description": "Description",
        "is_managed_externally": "Is Managed Externally",
        "owners.first_name": "Owners First Name",
        "owners.id": "Owners Id",
        "owners.last_name": "Owners Last Name",
        "owners.username": "Owners Username",
        "params": "Params",
        "query_context": "Query Context",
        "slice_name": "Slice Name",
        "viz_type": "Viz Type",
    },
    "result": {
        "cache_timeout": None,
        "certification_details": None,
        "certified_by": None,
        "dashboards": [{"dashboard_title": "Test", "id": 1, "json_metadata": None}],
        "description": None,
        "is_managed_externally": False,
        "owners": [
            {"first_name": "admin", "id": 1, "last_name": "admin", "username": "admin"},
        ],
        "params": json.dumps(
            {
                "adhoc_filters": [],
                "all_columns": ["number_of_orders"],
                "color_pn": True,
                "datasource": "1__table",
                "extra_form_data": {},
                "granularity_sqla": "first_order_date",
                "groupby": [],
                "order_by_cols": [],
                "order_desc": True,
                "percent_metrics": [],
                "query_mode": "raw",
                "row_limit": 10000,
                "server_page_length": 10,
                "show_cell_bars": True,
                "table_timestamp_format": "smart_date",
                "time_grain_sqla": "P1D",
                "time_range": "No filter",
                "viz_type": "table",
            },
        ),
        "query_context": json.dumps(
            {
                "datasource": {"id": 1, "type": "table"},
                "force": False,
                "queries": [
                    {
                        "time_range": "No filter",
                        "granularity": "first_order_date",
                        "filters": [],
                        "extras": {
                            "time_grain_sqla": "P1D",
                            "having": "",
                            "having_druid": [],
                            "where": "",
                        },
                        "applied_time_extras": {},
                        "columns": ["number_of_orders"],
                        "orderby": [],
                        "annotation_layers": [],
                        "row_limit": 10000,
                        "timeseries_limit": 0,
                        "order_desc": True,
                        "url_params": {},
                        "custom_params": {},
                        "custom_form_data": {},
                        "post_processing": [],
                    },
                ],
                "form_data": {
                    "viz_type": "table",
                    "datasource": "1__table",
                    "granularity_sqla": "first_order_date",
                    "time_grain_sqla": "P1D",
                    "time_range": "No filter",
                    "query_mode": "raw",
                    "groupby": [],
                    "all_columns": ["number_of_orders"],
                    "percent_metrics": [],
                    "adhoc_filters": [],
                    "order_by_cols": [],
                    "row_limit": 10000,
                    "server_page_length": 10,
                    "include_time": False,
                    "order_desc": True,
                    "table_timestamp_format": "smart_date",
                    "show_cell_bars": True,
                    "color_pn": True,
                    "extra_form_data": {},
                    "force": False,
                    "result_format": "json",
                    "result_type": "full",
                },
                "result_format": "json",
                "result_type": "full",
            },
        ),
        "slice_name": "Example chart",
        "viz_type": "pie",
    },
    "show_columns": [
        "cache_timeout",
        "certified_by",
        "certification_details",
        "dashboards.dashboard_title",
        "dashboards.id",
        "dashboards.json_metadata",
        "description",
        "owners.first_name",
        "owners.id",
        "owners.last_name",
        "owners.username",
        "params",
        "slice_name",
        "viz_type",
        "query_context",
        "is_managed_externally",
    ],
    "show_title": "Show Slice",
}


dataset_response = {
    "description_columns": {},
    "id": 27,
    "label_columns": {
        "cache_timeout": "Cache Timeout",
        "columns.changed_on": "Columns Changed On",
        "columns.column_name": "Columns Column Name",
        "columns.created_on": "Columns Created On",
        "columns.description": "Columns Description",
        "columns.expression": "Columns Expression",
        "columns.extra": "Columns Extra",
        "columns.filterable": "Columns Filterable",
        "columns.groupby": "Columns Groupby",
        "columns.id": "Columns Id",
        "columns.is_active": "Columns Is Active",
        "columns.is_dttm": "Columns Is Dttm",
        "columns.python_date_format": "Columns Python Date Format",
        "columns.type": "Columns Type",
        "columns.type_generic": "Columns Type Generic",
        "columns.uuid": "Columns Uuid",
        "columns.verbose_name": "Columns Verbose Name",
        "database.backend": "Database Backend",
        "database.database_name": "Database Database Name",
        "database.id": "Database Id",
        "datasource_type": "Datasource Type",
        "default_endpoint": "Default Endpoint",
        "description": "Description",
        "extra": "Extra",
        "fetch_values_predicate": "Fetch Values Predicate",
        "filter_select_enabled": "Filter Select Enabled",
        "id": "Id",
        "is_sqllab_view": "Is Sqllab View",
        "main_dttm_col": "Main Dttm Col",
        "metrics": "Metrics",
        "offset": "Offset",
        "owners.first_name": "Owners First Name",
        "owners.id": "Owners Id",
        "owners.last_name": "Owners Last Name",
        "owners.username": "Owners Username",
        "schema": "Schema",
        "sql": "Sql",
        "table_name": "Table Name",
        "template_params": "Template Params",
        "url": "Url",
    },
    "result": {
        "cache_timeout": None,
        "columns": [
            {
                "changed_on": "2022-03-27T13:21:33.957609",
                "column_name": "ts",
                "created_on": "2022-03-27T13:21:33.957602",
                "description": None,
                "expression": None,
                "extra": None,
                "filterable": True,
                "groupby": True,
                "id": 841,
                "is_active": True,
                "is_dttm": True,
                "python_date_format": None,
                "type": "TIMESTAMP WITHOUT TIME ZONE",
                "type_generic": 2,
                "uuid": "e607d7fd-90bf-4420-a35e-9dc7af555e0d",
                "verbose_name": None,
            },
            {
                "changed_on": "2022-03-27T13:21:33.958499",
                "column_name": "name",
                "created_on": "2022-03-27T13:21:33.958493",
                "description": None,
                "expression": None,
                "extra": None,
                "filterable": True,
                "groupby": True,
                "id": 842,
                "is_active": True,
                "is_dttm": False,
                "python_date_format": None,
                "type": "VARCHAR(255)",
                "type_generic": 1,
                "uuid": "76a523f0-1aad-4608-87a4-daf22172e1da",
                "verbose_name": None,
            },
            {
                "changed_on": "2022-03-27T13:21:33.975750",
                "column_name": "text",
                "created_on": "2022-03-27T13:21:33.975743",
                "description": None,
                "expression": None,
                "extra": None,
                "filterable": True,
                "groupby": True,
                "id": 843,
                "is_active": True,
                "is_dttm": False,
                "python_date_format": None,
                "type": "TEXT",
                "type_generic": 1,
                "uuid": "610b91b0-e8de-4703-bbe9-5b27fb6c6a4e",
                "verbose_name": None,
            },
        ],
        "database": {
            "backend": "postgresql",
            "database_name": "superset_examples_dev",
            "id": 3,
        },
        "datasource_type": "table",
        "default_endpoint": None,
        "description": "",
        "extra": json.dumps(
            {
                "resource_type": "model",
                "unique_id": "model.superset_examples.messages_channels",
                "depends_on": "ref('messages_channels')",
            },
        ),
        "fetch_values_predicate": None,
        "filter_select_enabled": False,
        "id": 27,
        "is_sqllab_view": False,
        "main_dttm_col": "ts",
        "metrics": [
            {
                "changed_on": "2022-03-27T13:21:34.298657",
                "created_on": "2022-03-27T13:21:34.248023",
                "d3format": None,
                "description": "",
                "expression": "count(*)",
                "extra": None,
                "id": 35,
                "metric_name": "cnt",
                "metric_type": "count",
                "uuid": "c4b74ceb-a19c-494a-9b90-9c68ed5bc8cb",
                "verbose_name": "count(*)",
                "warning_text": None,
            },
        ],
        "offset": 0,
        "owners": [],
        "schema": "public",
        "sql": None,
        "table_name": "messages_channels",
        "template_params": None,
        "url": "/tablemodelview/edit/27",
    },
    "show_columns": [
        "id",
        "database.database_name",
        "database.id",
        "table_name",
        "sql",
        "filter_select_enabled",
        "fetch_values_predicate",
        "schema",
        "description",
        "main_dttm_col",
        "offset",
        "default_endpoint",
        "cache_timeout",
        "is_sqllab_view",
        "template_params",
        "owners.id",
        "owners.username",
        "owners.first_name",
        "owners.last_name",
        "columns.changed_on",
        "columns.column_name",
        "columns.created_on",
        "columns.description",
        "columns.expression",
        "columns.filterable",
        "columns.groupby",
        "columns.id",
        "columns.is_active",
        "columns.extra",
        "columns.is_dttm",
        "columns.python_date_format",
        "columns.type",
        "columns.uuid",
        "columns.verbose_name",
        "metrics",
        "datasource_type",
        "url",
        "extra",
        "columns.type_generic",
        "database.backend",
    ],
    "show_title": "Show Sqla Table",
}

related_objects_response = {
    "charts": {
        "count": 1,
        "result": [{"id": 1, "slice_name": "Example chart", "viz_type": "pie"}],
    },
    "dashboards": {
        "count": 1,
        "result": [
            {
                "id": 12,
                "json_metadata": None,
                "slug": None,
                "title": "Example dashboard",
            },
        ],
    },
}


def test_get_dashboard_depends_on(mocker: MockerFixture) -> None:
    """
    Test ``get_dashboard_depends_on``.
    """
    client = mocker.MagicMock()
    client.get_dataset.return_value = dataset_response
    session = client.auth.get_session()
    session.get().json.return_value = datasets_response

    depends_on = get_dashboard_depends_on(client, dashboard_response["result"])
    assert depends_on == ["ref('messages_channels')"]


def test_get_dashboard_depends_on_no_extra(mocker: MockerFixture) -> None:
    """
    Test ``get_dashboard_depends_on`` without an extra payload.
    """
    client = mocker.MagicMock()
    modified_dataset_response = copy.deepcopy(dataset_response)
    modified_dataset_response["result"]["extra"] = None  # type: ignore
    client.get_dataset.return_value = modified_dataset_response
    session = client.auth.get_session()
    session.get().json.return_value = datasets_response

    depends_on = get_dashboard_depends_on(client, dashboard_response["result"])
    assert not depends_on


def test_get_dashboard_depends_on_invalid_extra(mocker: MockerFixture) -> None:
    """
    Test ``get_dashboard_depends_on`` when the extra payload is not JSON.
    """
    client = mocker.MagicMock()
    modified_dataset_response = copy.deepcopy(dataset_response)
    modified_dataset_response["result"]["extra"] = "{[("  # type: ignore
    client.get_dataset.return_value = modified_dataset_response
    session = client.auth.get_session()
    session.get().json.return_value = datasets_response

    depends_on = get_dashboard_depends_on(client, dashboard_response["result"])
    assert not depends_on


def test_get_chart_depends_on(mocker: MockerFixture) -> None:
    """
    Test ``get_chart_depends_on``.
    """
    client = mocker.MagicMock()
    client.get_dataset.return_value = dataset_response

    depends_on = get_chart_depends_on(client, chart_response["result"])
    assert depends_on == ["ref('messages_channels')"]


def test_get_chart_depends_on_no_extra(mocker: MockerFixture) -> None:
    """
    Test ``get_chart_depends_on`` without an extra payload.
    """
    client = mocker.MagicMock()
    modified_dataset_response = copy.deepcopy(dataset_response)
    modified_dataset_response["result"]["extra"] = None  # type: ignore
    client.get_dataset.return_value = modified_dataset_response

    depends_on = get_chart_depends_on(client, chart_response["result"])
    assert not depends_on


def test_sync_exposures(mocker: MockerFixture, fs: FakeFilesystem) -> None:
    """
    Test ``sync_exposures``.
    """
    root = Path("/path/to/root")
    fs.create_dir(root / "models")
    exposures = root / "models/exposures.yml"

    client = mocker.MagicMock()
    client.baseurl = URL("https://superset.example.org/")
    client.get_chart.return_value = chart_response
    client.get_dashboard.return_value = dashboard_response
    session = client.auth.get_session()
    session.get().json.return_value = related_objects_response
    mocker.patch(
        "preset_cli.cli.superset.sync.dbt.exposures.get_dashboard_depends_on",
        return_value=["ref('messages_channels')"],
    )
    mocker.patch(
        "preset_cli.cli.superset.sync.dbt.exposures.get_chart_depends_on",
        return_value=["ref('messages_channels')"],
    )

    datasets = [dataset_response["result"]]
    sync_exposures(client, exposures, datasets)

    with open(exposures, encoding="utf-8") as input_:
        contents = yaml.load(input_, Loader=yaml.SafeLoader)
    assert contents == {
        "version": 2,
        "exposures": [
            {
                "name": "Example chart [chart]",
                "type": "analysis",
                "maturity": "low",
                "url": (
                    "https://superset.example.org/superset/explore/"
                    "?form_data=%7B%22slice_id%22:+1%7D"
                ),
                "description": "",
                "depends_on": ["ref('messages_channels')"],
                "owner": {"name": "admin admin", "email": "unknown"},
            },
            {
                "name": "Example dashboard [dashboard]",
                "type": "dashboard",
                "maturity": "low",
                "url": "https://superset.example.org/superset/dashboard/12/",
                "description": "",
                "depends_on": ["ref('messages_channels')"],
                "owner": {"name": "admin admin", "email": "unknown"},
            },
        ],
    }


def test_sync_exposures_no_charts_no_dashboards(
    mocker: MockerFixture,
    fs: FakeFilesystem,
) -> None:
    """
    Test ``sync_exposures`` when no dashboads use the datasets.
    """
    root = Path("/path/to/root")
    fs.create_dir(root / "models")
    exposures = root / "models/exposures.yml"

    client = mocker.MagicMock()
    client.baseurl = URL("https://superset.example.org/")
    session = client.auth.get_session()
    no_related_objects_response = copy.deepcopy(related_objects_response)
    no_related_objects_response["charts"]["result"] = []
    no_related_objects_response["dashboards"]["result"] = []
    session.get().json.return_value = no_related_objects_response

    datasets = [dataset_response["result"]]
    sync_exposures(client, exposures, datasets)

    with open(exposures, encoding="utf-8") as input_:
        contents = yaml.load(input_, Loader=yaml.SafeLoader)
    assert contents == {
        "version": 2,
        "exposures": [],
    }



"""
Tests for the native import command.
"""
# pylint: disable=redefined-outer-name, invalid-name

from pathlib import Path
from typing import List
from unittest import mock
from zipfile import ZipFile

import pytest
import yaml
from click.testing import CliRunner
from freezegun import freeze_time
from jinja2 import Template
from pyfakefs.fake_filesystem import FakeFilesystem
from pytest_mock import MockerFixture

from preset_cli.cli.superset.main import superset_cli
from preset_cli.cli.superset.sync.native.command import (
    import_resource,
    load_user_modules,
    prompt_for_passwords,
    raise_helper,
)
from preset_cli.exceptions import ErrorLevel, ErrorPayload, SupersetError


def test_prompt_for_passwords(mocker: MockerFixture) -> None:
    """
    Test ``prompt_for_passwords``.
    """
    getpass = mocker.patch("preset_cli.cli.superset.sync.native.command.getpass")

    config = {"sqlalchemy_uri": "postgresql://user:XXXXXXXXXX@host:5432/db"}
    path = Path("/path/to/root/databases/psql.yaml")
    prompt_for_passwords(path, config)

    getpass.getpass.assert_called_with(f"Please provide the password for {path}: ")

    config["password"] = "password123"
    getpass.reset_mock()
    prompt_for_passwords(path, config)
    getpass.getpass.assert_not_called()


def test_import_resource(mocker: MockerFixture) -> None:
    """
    Test ``import_resource``.
    """
    client = mocker.MagicMock()

    contents = {"bundle/databases/gsheets.yaml": "GSheets"}
    with freeze_time("2022-01-01T00:00:00Z"):
        import_resource("database", contents=contents, client=client, overwrite=False)

    call = client.import_zip.mock_calls[0]
    assert call.kwargs == {"overwrite": False}

    resource, buf = call.args
    assert resource == "database"
    with ZipFile(buf) as bundle:
        assert bundle.namelist() == [
            "bundle/databases/gsheets.yaml",
            "bundle/metadata.yaml",
        ]
        assert (
            bundle.read("bundle/metadata.yaml").decode()
            == "timestamp: '2022-01-01T00:00:00+00:00'\ntype: Database\nversion: 1.0.0\n"
        )
        assert bundle.read("bundle/databases/gsheets.yaml").decode() == "GSheets"


def test_import_resource_overwrite_needed(mocker: MockerFixture) -> None:
    """
    Test ``import_resource`` when an overwrite error is raised.
    """
    click = mocker.patch("preset_cli.cli.superset.sync.native.command.click")
    client = mocker.MagicMock()
    errors: List[ErrorPayload] = [
        {
            "message": "Error importing database",
            "error_type": "GENERIC_COMMAND_ERROR",
            "level": ErrorLevel.WARNING,
            "extra": {
                "databases/gsheets.yaml": (
                    "Database already exists and `overwrite=true` was not passed"
                ),
                "issue_codes": [
                    {
                        "code": 1010,
                        "message": (
                            "Issue 1010 - Superset encountered an "
                            "error while running a command."
                        ),
                    },
                ],
            },
        },
    ]
    client.import_zip.side_effect = SupersetError(errors)

    contents = {"bundle/databases/gsheets.yaml": "GSheets"}
    import_resource("database", contents=contents, client=client, overwrite=False)

    assert click.style.called_with(
        "The following file(s) already exist. Pass --overwrite to replace them.\n"
        "databases/gsheets.yaml",
        fg="bright_red",
    )


def test_import_resource_error(mocker: MockerFixture) -> None:
    """
    Test ``import_resource`` when an unexpected error is raised.
    """
    client = mocker.MagicMock()
    errors: List[ErrorPayload] = [
        {
            "message": "Error importing database",
            "error_type": "GENERIC_COMMAND_ERROR",
            "level": ErrorLevel.WARNING,
            "extra": {
                "issue_codes": [
                    {
                        "code": 1010,
                        "message": (
                            "Issue 1010 - Superset encountered an "
                            "error while running a command."
                        ),
                    },
                ],
            },
        },
    ]
    client.import_zip.side_effect = SupersetError(errors)

    contents = {"bundle/databases/gsheets.yaml": "GSheets"}
    with pytest.raises(SupersetError) as excinfo:
        import_resource("database", contents=contents, client=client, overwrite=False)
    assert excinfo.value.errors == errors


def test_native(mocker: MockerFixture, fs: FakeFilesystem) -> None:
    """
    Test the ``native`` command.
    """
    root = Path("/path/to/root")
    fs.create_dir(root)
    database_config = {
        "database_name": "GSheets",
        "sqlalchemy_uri": "gsheets://",
        "is_managed_externally": False,
    }
    dataset_config = {"table_name": "test", "is_managed_externally": False}
    fs.create_file(
        root / "databases/gsheets.yaml",
        contents=yaml.dump(database_config),
    )
    fs.create_file(
        root / "datasets/gsheets/test.yaml",
        contents=yaml.dump(dataset_config),
    )
    fs.create_file(
        root / "README.txt",
        contents="Hello, world",
    )

    SupersetClient = mocker.patch(
        "preset_cli.cli.superset.sync.native.command.SupersetClient",
    )
    client = SupersetClient()
    import_resource = mocker.patch(
        "preset_cli.cli.superset.sync.native.command.import_resource",
    )
    mocker.patch("preset_cli.cli.superset.main.UsernamePasswordAuth")

    runner = CliRunner()
    result = runner.invoke(
        superset_cli,
        ["https://superset.example.org/", "sync", "native", str(root)],
        catch_exceptions=False,
    )
    assert result.exit_code == 0
    contents = {
        "bundle/databases/gsheets.yaml": yaml.dump(database_config),
        "bundle/datasets/gsheets/test.yaml": yaml.dump(dataset_config),
    }
    import_resource.assert_has_calls(
        [
            mock.call("database", contents, client, False),
            mock.call("dataset", contents, client, False),
            mock.call("chart", contents, client, False),
            mock.call("dashboard", contents, client, False),
        ],
    )


def test_native_load_env(
    mocker: MockerFixture,
    fs: FakeFilesystem,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """
    Test the ``native`` command with the ``--load-env`` option.
    """
    monkeypatch.setenv("SQLALCHEMY_URI", "postgres://host1")

    root = Path("/path/to/root")
    fs.create_dir(root)
    database_config = {
        "database_name": "Postgres",
        "sqlalchemy_uri": '{{ env["SQLALCHEMY_URI"] }}',
        "is_managed_externally": False,
    }
    fs.create_file(
        root / "databases/postgres.yaml",
        contents=yaml.dump(database_config),
    )

    SupersetClient = mocker.patch(
        "preset_cli.cli.superset.sync.native.command.SupersetClient",
    )
    client = SupersetClient()
    import_resource = mocker.patch(
        "preset_cli.cli.superset.sync.native.command.import_resource",
    )
    mocker.patch("preset_cli.cli.superset.main.UsernamePasswordAuth")

    runner = CliRunner()
    result = runner.invoke(
        superset_cli,
        [
            "https://superset.example.org/",
            "sync",
            "native",
            str(root),
            "-e",
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 0
    contents = {
        "bundle/databases/postgres.yaml": yaml.dump(
            {
                "database_name": "Postgres",
                "sqlalchemy_uri": "postgres://host1",
                "is_managed_externally": False,
            },
        ),
    }
    print(import_resource.mock_calls)
    import_resource.assert_has_calls(
        [
            mock.call("database", contents, client, False),
            mock.call("dataset", contents, client, False),
            mock.call("chart", contents, client, False),
            mock.call("dashboard", contents, client, False),
        ],
    )


def test_native_external_url(mocker: MockerFixture, fs: FakeFilesystem) -> None:
    """
    Test the ``native`` command with an external URL.
    """
    root = Path("/path/to/root")
    fs.create_dir(root)
    database_config = {
        "database_name": "GSheets",
        "sqlalchemy_uri": "gsheets://",
        "external_url": "https://repo.example.com/databases/gsheets.yaml",
        "is_managed_externally": True,
    }
    dataset_config = {
        "table_name": "test",
        "external_url": "https://repo.example.com/datasets/gsheets/test.yaml",
        "is_managed_externally": True,
    }
    fs.create_file(
        root / "databases/gsheets.yaml",
        contents=yaml.dump(database_config),
    )
    fs.create_file(
        root / "datasets/gsheets/test.yaml",
        contents=yaml.dump(dataset_config),
    )

    SupersetClient = mocker.patch(
        "preset_cli.cli.superset.sync.native.command.SupersetClient",
    )
    client = SupersetClient()
    import_resource = mocker.patch(
        "preset_cli.cli.superset.sync.native.command.import_resource",
    )
    mocker.patch("preset_cli.cli.superset.main.UsernamePasswordAuth")

    runner = CliRunner()
    result = runner.invoke(
        superset_cli,
        [
            "https://superset.example.org/",
            "sync",
            "native",
            str(root),
            "--external-url-prefix",
            "https://repo.example.com/",
            "--disallow-edits",
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 0
    database_config["external_url"] = "https://repo.example.com/databases/gsheets.yaml"
    dataset_config[
        "external_url"
    ] = "https://repo.example.com/datasets/gsheets/test.yaml"
    contents = {
        "bundle/databases/gsheets.yaml": yaml.dump(database_config),
        "bundle/datasets/gsheets/test.yaml": yaml.dump(dataset_config),
    }
    import_resource.assert_has_calls(
        [
            mock.call("database", contents, client, False),
            mock.call("dataset", contents, client, False),
            mock.call("chart", contents, client, False),
            mock.call("dashboard", contents, client, False),
        ],
    )


def test_load_user_modules(mocker: MockerFixture, fs: FakeFilesystem) -> None:
    """
    Test ``load_user_modules``.
    """
    importlib = mocker.patch("preset_cli.cli.superset.sync.native.command.importlib")
    spec = mocker.MagicMock()
    importlib.util.spec_from_file_location.side_effect = [spec, None]

    root = Path("/path/to/root")
    fs.create_dir(root)
    fs.create_file(root / "test.py")
    fs.create_file(root / "invalid.py")
    fs.create_file(root / "test.txt")

    load_user_modules(root)

    importlib.util.spec_from_file_location.assert_has_calls(
        [
            mock.call("test", root / "test.py"),
            mock.call("invalid", root / "invalid.py"),
        ],
    )
    assert importlib.util.module_from_spec.call_count == 1


def test_raise_helper() -> None:
    """
    Test the ``raise_helper`` macro.
    """
    template = Template(
        """
{% if i > 0  %}
{{ i }}
{% else %}
{{ raise("Invalid number: %d", i) }}
{% endif %}
    """,
    )

    assert template.render(i=1, **{"raise": raise_helper}) == "\n\n1\n\n    "
    with pytest.raises(Exception) as excinfo:
        template.render(i=-1, **{"raise": raise_helper})
    assert str(excinfo.value) == "Invalid number: -1"



"""
Sync dbt database to Superset.
"""

import logging
from pathlib import Path
from typing import Any

from yarl import URL

from preset_cli.api.clients.superset import SupersetClient
from preset_cli.cli.superset.sync.dbt.lib import build_sqlalchemy_params, load_profiles
from preset_cli.exceptions import DatabaseNotFoundError

_logger = logging.getLogger(__name__)


def sync_database(  # pylint: disable=too-many-locals, too-many-arguments
    client: SupersetClient,
    profiles_path: Path,
    project_name: str,
    target_name: str,
    import_db: bool,
    disallow_edits: bool,  # pylint: disable=unused-argument
    external_url_prefix: str,
) -> Any:
    """
    Read target database from a dbt profiles.yml and sync to Superset.
    """
    base_url = URL(external_url_prefix) if external_url_prefix else None

    profiles = load_profiles(profiles_path, project_name, target_name)
    project = profiles[project_name]
    outputs = project["outputs"]
    target = outputs[target_name]

    # read additional metadata that should be applied to the DB
    meta = target.get("meta", {}).get("superset", {})

    if "connection_params" in meta:
        connection_params = meta.pop("connection_params")
    else:
        connection_params = build_sqlalchemy_params(target)

    database_name = meta.pop("database_name", f"{project_name}_{target_name}")
    databases = client.get_databases(database_name=database_name)
    if len(databases) > 1:
        raise Exception("More than one database with the same name found")

    if base_url and "external_url" not in meta:
        meta["external_url"] = str(base_url.with_fragment("!/overview"))

    if databases:
        _logger.info("Found an existing database, updating it")
        database = databases[0]
        database = client.update_database(
            database_id=database["id"],
            database_name=database_name,
            is_managed_externally=disallow_edits,
            **meta,
        )
    elif not import_db:
        raise DatabaseNotFoundError()
    else:
        _logger.info("No database found, creating it")
        database = client.create_database(
            database_name=database_name,
            is_managed_externally=disallow_edits,
            **connection_params,
            **meta,
        )

    return database



"""
Tests for the export commands.
"""
# pylint: disable=redefined-outer-name, invalid-name, unused-argument

from io import BytesIO
from pathlib import Path
from unittest import mock
from uuid import UUID
from zipfile import ZipFile

import pytest
import yaml
from click.testing import CliRunner
from pyfakefs.fake_filesystem import FakeFilesystem
from pytest_mock import MockerFixture

from preset_cli.auth.main import Auth
from preset_cli.cli.superset.export import export_resource
from preset_cli.cli.superset.main import superset_cli


@pytest.fixture
def dataset_export() -> BytesIO:
    """
    Fixture for the contents of a simple database export.
    """
    contents = {
        "dashboard_export/metadata.yaml": "Metadata",
        "dashboard_export/databases/gsheets.yaml": yaml.dump(
            {
                "database_name": "GSheets",
                "sqlalchemy_uri": "gsheets://",
            },
        ),
        "dashboard_export/datasets/gsheets/test.yaml": yaml.dump(
            {
                "table_name": "test",
                "sql": """
SELECT action, count(*) as times
FROM logs
WHERE
    action in {{ filter_values('action_type')|where_in }}
GROUP BY action""",
            },
        ),
    }

    buf = BytesIO()
    with ZipFile(buf, "w") as bundle:
        for file_name, file_contents in contents.items():
            with bundle.open(file_name, "w") as output:
                output.write(file_contents.encode())
    buf.seek(0)
    return buf


def test_export_resource(
    mocker: MockerFixture,
    fs: FakeFilesystem,
    dataset_export: BytesIO,
) -> None:
    """
    Test ``export_resource``.
    """
    root = Path("/path/to/root")
    fs.create_dir(root)

    client = mocker.MagicMock()
    client.export_zip.return_value = dataset_export

    export_resource(resource_name="database", root=root, client=client, overwrite=False)
    with open(root / "databases/gsheets.yaml", encoding="utf-8") as input_:
        assert input_.read() == "database_name: GSheets\nsqlalchemy_uri: gsheets://\n"

    # check that Jinja2 was escaped
    export_resource(resource_name="dataset", root=root, client=client, overwrite=False)
    with open(root / "datasets/gsheets/test.yaml", encoding="utf-8") as input_:
        assert yaml.load(input_.read(), Loader=yaml.SafeLoader) == {
            "table_name": "test",
            "sql": """
SELECT action, count(*) as times
FROM logs
WHERE
    action in {{ '{{' }} filter_values('action_type')|where_in {{ '}}' }}
GROUP BY action""",
        }

    # metadata file should be ignored
    assert not (root / "metadata.yaml").exists()


def test_export_resource_overwrite(
    mocker: MockerFixture,
    fs: FakeFilesystem,
    dataset_export: BytesIO,
) -> None:
    """
    Test that we need to confirm overwrites.
    """
    root = Path("/path/to/root")
    fs.create_dir(root)

    client = mocker.MagicMock()
    client.export_zip.return_value = dataset_export

    export_resource(resource_name="database", root=root, client=client, overwrite=False)
    with pytest.raises(Exception) as excinfo:
        export_resource(
            resource_name="database",
            root=root,
            client=client,
            overwrite=False,
        )
    assert str(excinfo.value) == (
        "File already exists and --overwrite was not specified: "
        "/path/to/root/databases/gsheets.yaml"
    )

    export_resource(resource_name="database", root=root, client=client, overwrite=True)


def test_export_assets(mocker: MockerFixture, fs: FakeFilesystem) -> None:
    """
    Test the ``export_assets`` command.
    """
    # root must exist for command to succeed
    root = Path("/path/to/root")
    fs.create_dir(root)

    SupersetClient = mocker.patch("preset_cli.cli.superset.export.SupersetClient")
    client = SupersetClient()
    export_resource = mocker.patch("preset_cli.cli.superset.export.export_resource")
    mocker.patch("preset_cli.cli.superset.main.UsernamePasswordAuth")

    runner = CliRunner()
    result = runner.invoke(
        superset_cli,
        ["https://superset.example.org/", "export", "/path/to/root"],
        catch_exceptions=False,
    )
    assert result.exit_code == 0
    export_resource.assert_has_calls(
        [
            mock.call("database", Path("/path/to/root"), client, False),
            mock.call("dataset", Path("/path/to/root"), client, False),
            mock.call("chart", Path("/path/to/root"), client, False),
            mock.call("dashboard", Path("/path/to/root"), client, False),
        ],
    )


def test_export_with_custom_auth(mocker: MockerFixture, fs: FakeFilesystem) -> None:
    """
    Test the ``export`` command.
    """
    # root must exist for command to succeed
    root = Path("/path/to/root")
    fs.create_dir(root)

    SupersetClient = mocker.patch("preset_cli.cli.superset.export.SupersetClient")
    client = SupersetClient()
    export_resource = mocker.patch("preset_cli.cli.superset.export.export_resource")

    runner = CliRunner()
    result = runner.invoke(
        superset_cli,
        ["https://superset.example.org/", "export", "/path/to/root"],
        catch_exceptions=False,
        obj={"AUTH": Auth()},
    )
    assert result.exit_code == 0
    export_resource.assert_has_calls(
        [
            mock.call("database", Path("/path/to/root"), client, False),
            mock.call("dataset", Path("/path/to/root"), client, False),
            mock.call("chart", Path("/path/to/root"), client, False),
            mock.call("dashboard", Path("/path/to/root"), client, False),
        ],
    )


def test_export_users(mocker: MockerFixture, fs: FakeFilesystem) -> None:
    """
    Test the ``export_users`` command.
    """
    mocker.patch("preset_cli.cli.superset.main.UsernamePasswordAuth")
    SupersetClient = mocker.patch("preset_cli.cli.superset.export.SupersetClient")
    client = SupersetClient()
    client.export_users.return_value = [
        {
            "first_name": "admin",
            "last_name": "admin",
            "username": "admin",
            "email": "admin@example.com",
            "role": ["Admin"],
        },
    ]

    runner = CliRunner()
    result = runner.invoke(
        superset_cli,
        ["https://superset.example.org/", "export-users", "users.yaml"],
        catch_exceptions=False,
    )
    assert result.exit_code == 0

    with open("users.yaml", encoding="utf-8") as input_:
        contents = yaml.load(input_, Loader=yaml.SafeLoader)
    print(contents)
    assert contents == [
        {
            "first_name": "admin",
            "last_name": "admin",
            "username": "admin",
            "email": "admin@example.com",
            "role": ["Admin"],
        },
    ]


def test_export_rls(mocker: MockerFixture, fs: FakeFilesystem) -> None:
    """
    Test the ``export_users`` command.
    """
    mocker.patch("preset_cli.cli.superset.main.UsernamePasswordAuth")
    SupersetClient = mocker.patch("preset_cli.cli.superset.export.SupersetClient")
    client = SupersetClient()
    client.export_rls.return_value = [
        {
            "clause": "client_id = 9",
            "description": "This is a rule. There are many others like it, but this one is mine.",
            "filter_type": "Regular",
            "group_key": "department",
            "name": "My rule",
            "roles": ["Gamma"],
            "tables": ["main.test_table"],
        },
    ]

    runner = CliRunner()
    result = runner.invoke(
        superset_cli,
        ["https://superset.example.org/", "export-rls", "rls.yaml"],
        catch_exceptions=False,
    )
    assert result.exit_code == 0

    with open("rls.yaml", encoding="utf-8") as input_:
        contents = yaml.load(input_, Loader=yaml.SafeLoader)
    assert contents == [
        {
            "clause": "client_id = 9",
            "description": "This is a rule. There are many others like it, but this one is mine.",
            "filter_type": "Regular",
            "group_key": "department",
            "name": "My rule",
            "roles": ["Gamma"],
            "tables": ["main.test_table"],
        },
    ]


def test_export_ownership(mocker: MockerFixture, fs: FakeFilesystem) -> None:
    """
    Test the ``export_ownership`` command.
    """
    mocker.patch("preset_cli.cli.superset.main.UsernamePasswordAuth")
    SupersetClient = mocker.patch("preset_cli.cli.superset.export.SupersetClient")
    client = SupersetClient()
    client.export_ownership.side_effect = [
        [],
        [
            {
                "name": "My chart",
                "uuid": UUID("e0d20af0-cef9-4bdb-80b4-745827f441bf"),
                "owners": ["adoe@example.com", "bdoe@example.com"],
            },
        ],
        [],
    ]

    runner = CliRunner()
    result = runner.invoke(
        superset_cli,
        ["https://superset.example.org/", "export-ownership"],
        catch_exceptions=False,
    )
    assert result.exit_code == 0

    with open("ownership.yaml", encoding="utf-8") as input_:
        contents = yaml.load(input_, Loader=yaml.SafeLoader)
    assert contents == {
        "chart": [
            {
                "name": "My chart",
                "uuid": "e0d20af0-cef9-4bdb-80b4-745827f441bf",
                "owners": ["adoe@example.com", "bdoe@example.com"],
            },
        ],
    }



"""
Mechanisms for authentication and authorization.
"""

from typing import Dict, Optional

import requests
from bs4 import BeautifulSoup
from yarl import URL


class Auth:  # pylint: disable=too-few-public-methods
    """
    An authentication/authorization mechanism.
    """

    def __init__(self):
        self.session = requests.Session()
        self.headers = {}

    def get_session(self) -> requests.Session:
        """
        Return a session.
        """
        return self.session

    def get_headers(self) -> Dict[str, str]:  # pylint: disable=no-self-use
        """
        Return headers for auth.
        """
        return self.headers


class UsernamePasswordAuth(Auth):  # pylint: disable=too-few-public-methods
    """
    Auth via username/password.
    """

    def __init__(self, baseurl: URL, username: str, password: Optional[str] = None):
        super().__init__()
        self._do_login(baseurl, username, password)

    def _do_login(
        self,
        baseurl: URL,
        username: str,
        password: Optional[str] = None,
    ) -> None:
        """
        Login to get CSRF token and cookies.
        """
        response = self.session.get(baseurl / "login/")
        soup = BeautifulSoup(response.text, "html.parser")
        csrf_token = soup.find("input", {"id": "csrf_token"})["value"]

        # update headers
        self.headers["X-CSRFToken"] = csrf_token

        # set cookies
        self.session.post(
            baseurl / "login/",
            data=dict(username=username, password=password, csrf_token=csrf_token),
        )



"""
Operators for filtering the API.
"""

# pylint: disable=too-few-public-methods

from typing import Any


class Operator:
    """
    A filter operator.
    """

    operator = "invalid"

    def __init__(self, value: Any):
        self.value = value


class Equal(Operator):
    """
    Equality operator.
    """

    operator = "eq"


class OneToMany(Operator):
    """
    Operator for one-to-many relationships.
    """

    operator = "rel_o_m"



"""
Metric conversion.

This module is used to convert dbt metrics into Superset metrics.
"""

# pylint: disable=consider-using-f-string

from functools import partial
from typing import Dict, List

from jinja2 import Template

from preset_cli.api.clients.dbt import FilterSchema, MetricSchema


def get_metric_expression(metric_name: str, metrics: Dict[str, MetricSchema]) -> str:
    """
    Return a SQL expression for a given dbt metric.
    """
    if metric_name not in metrics:
        raise Exception(f"Invalid metric {metric_name}")

    metric = metrics[metric_name]
    type_ = metric["type"]
    sql = metric["sql"]

    if metric.get("filters"):
        sql = apply_filters(sql, metric["filters"])

    simple_mappings = {
        "count": "COUNT",
        "sum": "SUM",
        "average": "AVG",
        "min": "MIN",
        "max": "MAX",
    }

    if type_ in simple_mappings:
        function = simple_mappings[type_]
        return f"{function}({sql})"

    if type_ == "count_distinct":
        return f"COUNT(DISTINCT {sql})"

    if type_ == "expression":
        template = Template(sql)
        return template.render(metric=partial(get_metric_expression, metrics=metrics))

    sorted_metric = dict(sorted(metric.items()))
    raise Exception(f"Unable to generate metric expression from: {sorted_metric}")


def apply_filters(sql: str, filters: List[FilterSchema]) -> str:
    """
    Apply filters to SQL expression.
    """
    condition = " AND ".join(
        "{field} {operator} {value}".format(**filter_) for filter_ in filters
    )
    return f"CASE WHEN {condition} THEN {sql} END"



"""
Sync Superset dashboards as dbt exposures.
"""

import json
from pathlib import Path
from typing import Any, List

import yaml

from preset_cli.api.clients.superset import SupersetClient

# XXX: DashboardResponseType and DatasetResponseType


def get_chart_depends_on(client: SupersetClient, chart: Any) -> List[str]:
    """
    Get all the dbt dependencies for a given chart.
    """

    query_context = json.loads(chart["query_context"])
    dataset_id = query_context["datasource"]["id"]
    dataset = client.get_dataset(dataset_id)
    extra = json.loads(dataset["result"]["extra"] or "{}")
    if "depends_on" in extra:
        return [extra["depends_on"]]

    return []


def get_dashboard_depends_on(client: SupersetClient, dashboard: Any) -> List[str]:
    """
    Get all the dbt dependencies for a given dashboard.
    """

    url = client.baseurl / "api/v1/dashboard" / str(dashboard["id"]) / "datasets"

    session = client.auth.get_session()
    headers = client.auth.get_headers()
    response = session.get(url, headers=headers)
    response.raise_for_status()

    payload = response.json()

    depends_on = []
    for dataset in payload["result"]:
        full_dataset = client.get_dataset(int(dataset["id"]))
        try:
            extra = json.loads(full_dataset["result"]["extra"] or "{}")
        except json.decoder.JSONDecodeError:
            extra = {}
        if "depends_on" in extra:
            depends_on.append(extra["depends_on"])

    return depends_on


def sync_exposures(  # pylint: disable=too-many-locals
    client: SupersetClient,
    exposures_path: Path,
    datasets: List[Any],
) -> None:
    """
    Write dashboards back to dbt as exposures.
    """
    exposures = []
    charts_ids = set()
    dashboards_ids = set()

    for dataset in datasets:
        url = client.baseurl / "api/v1/dataset" / str(dataset["id"]) / "related_objects"

        session = client.auth.get_session()
        headers = client.auth.get_headers()
        response = session.get(url, headers=headers)
        response.raise_for_status()

        payload = response.json()
        for chart in payload["charts"]["result"]:
            charts_ids.add(chart["id"])
        for dashboard in payload["dashboards"]["result"]:
            dashboards_ids.add(dashboard["id"])

    for chart_id in charts_ids:
        chart = client.get_chart(chart_id)["result"]
        first_owner = chart["owners"][0]
        exposure = {
            "name": chart["slice_name"] + " [chart]",
            "type": "analysis",
            "maturity": "high" if chart["certified_by"] else "low",
            "url": str(
                client.baseurl
                / "superset/explore/"
                % {"form_data": json.dumps({"slice_id": chart_id})},
            ),
            "description": chart["description"] or "",
            "depends_on": get_chart_depends_on(client, chart),
            "owner": {
                "name": first_owner["first_name"] + " " + first_owner["last_name"],
                "email": first_owner.get("email", "unknown"),
            },
        }
        exposures.append(exposure)

    for dashboard_id in dashboards_ids:
        dashboard = client.get_dashboard(dashboard_id)["result"]
        first_owner = dashboard["owners"][0]
        exposure = {
            "name": dashboard["dashboard_title"] + " [dashboard]",
            "type": "dashboard",
            "maturity": "high"
            if dashboard["published"] or dashboard["certified_by"]
            else "low",
            "url": str(client.baseurl / dashboard["url"].lstrip("/")),
            "description": "",
            "depends_on": get_dashboard_depends_on(client, dashboard),
            "owner": {
                "name": first_owner["first_name"] + " " + first_owner["last_name"],
                "email": first_owner.get("email", "unknown"),
            },
        }
        exposures.append(exposure)

    with open(exposures_path, "w", encoding="utf-8") as output:
        yaml.safe_dump({"version": 2, "exposures": exposures}, output, sort_keys=False)



"""
Basic helper functions.
"""

import json
import logging
from pathlib import Path
from typing import Any, Dict, List, cast

from requests import Response
from rich.logging import RichHandler

from preset_cli.exceptions import ErrorLevel, ErrorPayload, SupersetError

_logger = logging.getLogger(__name__)


def remove_root(file_path: str) -> str:
    """
    Remove the first directory of a path.
    """
    full_path = Path(file_path)
    return str(Path(*full_path.parts[1:]))


def setup_logging(loglevel: str) -> None:
    """
    Setup basic logging.
    """
    level = getattr(logging, loglevel.upper(), None)
    if not isinstance(level, int):
        raise ValueError(f"Invalid log level: {loglevel}")

    logformat = "[%(asctime)s] %(levelname)s: %(name)s: %(message)s"
    logging.basicConfig(
        level=level,
        format=logformat,
        datefmt="[%X]",
        handlers=[RichHandler()],
        force=True,
    )


def deserialize_error_level(errors: List[Dict[str, Any]]) -> List[ErrorPayload]:
    """
    Convert error level from string to enum.
    """
    for error in errors:
        if isinstance(error, dict) and isinstance(error.get("level"), str):
            error["level"] = ErrorLevel(error["level"])
    return cast(List[ErrorPayload], errors)


def is_sip_40_payload(errors: List[Dict[str, Any]]) -> bool:
    """
    Return if a given error payload comforms with SIP-40.
    """
    return isinstance(errors, list) and all(
        isinstance(error, dict)
        and set(error.keys()) <= {"message", "error_type", "level", "extra"}
        for error in errors
    )


def validate_response(response: Response) -> None:
    """
    Check for errors in a response.
    """
    if response.ok:
        return

    if response.headers.get("content-type") == "application/json":
        payload = response.json()
        message = json.dumps(payload, indent=4)

        if "errors" in payload and is_sip_40_payload(payload["errors"]):
            errors = deserialize_error_level(payload["errors"])
        else:
            errors = [
                {
                    "message": "Unknown error",
                    "error_type": "UNKNOWN_ERROR",
                    "level": ErrorLevel.ERROR,
                    "extra": payload,
                },
            ]
    else:
        message = response.text
        errors = [
            {
                "message": message,
                "error_type": "UNKNOWN_ERROR",
                "level": ErrorLevel.ERROR,
            },
        ]

    _logger.error(message)
    raise SupersetError(errors=errors)



"""
Tests for metrics.
"""

from typing import Dict

import pytest

from preset_cli.api.clients.dbt import MetricSchema
from preset_cli.cli.superset.sync.dbt.metrics import get_metric_expression


def test_get_metric_expression() -> None:
    """
    Tests for ``get_metric_expression``.
    """
    metric_schema = MetricSchema()
    metrics: Dict[str, MetricSchema] = {
        "one": metric_schema.load(
            {
                "type": "count",
                "sql": "user_id",
                "filters": [
                    {"field": "is_paying", "operator": "is", "value": "true"},
                    {"field": "lifetime_value", "operator": ">=", "value": "100"},
                    {"field": "company_name", "operator": "!=", "value": "'Acme, Inc'"},
                    {"field": "signup_date", "operator": ">=", "value": "'2020-01-01'"},
                ],
            },
        ),
        "two": metric_schema.load(
            {
                "type": "count_distinct",
                "sql": "user_id",
            },
        ),
        "three": metric_schema.load(
            {
                "type": "expression",
                "sql": "{{metric('one')}} - {{metric('two')}}",
            },
        ),
        "four": metric_schema.load(
            {
                "type": "hllsketch",
                "sql": "user_id",
            },
        ),
    }
    assert get_metric_expression("one", metrics) == (
        "COUNT(CASE WHEN is_paying is true AND lifetime_value >= 100 AND "
        "company_name != 'Acme, Inc' AND signup_date >= '2020-01-01' THEN user_id END)"
    )

    assert get_metric_expression("two", metrics) == "COUNT(DISTINCT user_id)"

    assert get_metric_expression("three", metrics) == (
        "COUNT(CASE WHEN is_paying is true AND lifetime_value >= 100 AND "
        "company_name != 'Acme, Inc' AND signup_date >= '2020-01-01' THEN user_id END) "
        "- COUNT(DISTINCT user_id)"
    )

    with pytest.raises(Exception) as excinfo:
        get_metric_expression("four", metrics)
    assert str(excinfo.value) == (
        "Unable to generate metric expression from: "
        "{'sql': 'user_id', 'type': 'hllsketch'}"
    )

    with pytest.raises(Exception) as excinfo:
        get_metric_expression("five", metrics)
    assert str(excinfo.value) == "Invalid metric five"
